@@ -6,22 +6,26 @@
 //! [`Write`]: https://doc.rust-lang.org/stable/std/io/trait.Write.html
 
 use crate::{
-    BlockSize, CompressionMode, IntoInnerError, LargeWindowSize, SetParameterError, Quality,
+    BlockSize, CompressionMode, IntoInnerError, LargeWindowSize, Quality, SetParameterError,
     WindowSize,
 };
 use brotlic_sys::*;
+use std::alloc::GlobalAlloc;
+use std::error::Error;
 use std::io::{BufRead, Read, Write};
 use std::{fmt, io, mem, ptr, slice};
-use std::error::Error;
 
 /// A reference to a brotli encoder.
 ///
 /// This encoder contains internal state of the encoding process. This low-level wrapper intended to
 /// be used for people who are familiar with the C API. For higher level abstractions, see
 /// [`CompressorReader`] and [`CompressorWriter`].
-#[derive(Debug)]
 pub struct BrotliEncoder {
     state: *mut BrotliEncoderState,
+
+    // this field is read read across FFI boundaries
+    #[allow(dead_code)]
+    alloc: Option<Box<Box<dyn GlobalAlloc>>>,
 }
 
 unsafe impl Send for BrotliEncoder {}
@@ -35,16 +39,41 @@ impl BrotliEncoder {
     /// Panics if the encoder fails to be allocated or initialized
     #[doc(alias = "BrotliEncoderCreateInstance")]
     pub fn new() -> Self {
-        unsafe {
-            let instance = BrotliEncoderCreateInstance(None, None, ptr::null_mut());
+        let instance = unsafe { BrotliEncoderCreateInstance(None, None, ptr::null_mut()) };
 
-            if !instance.is_null() {
-                BrotliEncoder { state: instance }
-            } else {
-                panic!(
-                    "BrotliEncoderCreateInstance returned NULL: failed to allocate or initialize"
-                );
+        if !instance.is_null() {
+            BrotliEncoder {
+                state: instance,
+                alloc: None,
             }
+        } else {
+            panic!("BrotliEncoderCreateInstance returned NULL: failed to allocate or initialize");
+        }
+    }
+
+    /// Constructs a new brotli encoder instance using allocator `alloc`.
+    ///
+    /// # Panics
+    ///
+    /// Panics if the encoder fails to be allocated or initialized
+    #[doc(alias = "BrotliEncoderCreateInstance")]
+    pub fn new_in<A>(alloc: A) -> Self
+    where
+        A: GlobalAlloc + 'static,
+    {
+        let alloc: Box<Box<dyn GlobalAlloc>> = Box::new(Box::new(alloc));
+        let alloc_ptr: *const Box<dyn GlobalAlloc> = alloc.as_ref();
+        let instance = unsafe {
+            BrotliEncoderCreateInstance(Some(crate::malloc), Some(crate::free), alloc_ptr as _)
+        };
+
+        if !instance.is_null() {
+            BrotliEncoder {
+                state: instance,
+                alloc: Some(alloc),
+            }
+        } else {
+            panic!("BrotliEncoderCreateInstance returned NULL: failed to allocate or initialize");
         }
     }
 
@@ -140,6 +169,7 @@ impl BrotliEncoder {
     }
 
     /// Checks if the encoder has more output.
+    #[doc(alias = "BrotliEncoderHasMoreOutput")]
     pub fn has_output(&self) -> bool {
         unsafe { BrotliEncoderHasMoreOutput(self.state) != 0 }
     }
@@ -155,7 +185,6 @@ impl BrotliEncoder {
     /// # Safety
     ///
     /// For every consecutive call of this function, the previous slice becomes invalidated.
-    #[doc(alias = "BrotliEncoderHasMoreOutput")]
     #[doc(alias = "BrotliEncoderTakeOutput")]
     pub unsafe fn take_output(&mut self) -> Option<&[u8]> {
         if self.has_output() {
@@ -169,6 +198,7 @@ impl BrotliEncoder {
     }
 
     /// Returns the version of the C brotli encoder library.
+    #[doc(alias = "BrotliEncoderVersion")]
     pub fn version() -> u32 {
         unsafe { BrotliEncoderVersion() }
     }
@@ -193,6 +223,14 @@ impl BrotliEncoder {
     }
 }
 
+impl fmt::Debug for BrotliEncoder {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("BrotliEncoder")
+            .field("state", &self.state)
+            .finish_non_exhaustive()
+    }
+}
+
 impl Default for BrotliEncoder {
     fn default() -> Self {
         BrotliEncoder::new()
@@ -371,7 +409,7 @@ impl BrotliEncoderOptions {
         self
     }
 
-    /// Creates a brotli encoder using the specified settings.
+    /// Creates a brotli encoder with the specified settings using allocator `alloc`.
     ///
     /// # Errors
     ///
@@ -380,6 +418,29 @@ impl BrotliEncoderOptions {
     pub fn build(&self) -> Result<BrotliEncoder, SetParameterError> {
         let mut encoder = BrotliEncoder::new();
 
+        self.configure(&mut encoder)?;
+
+        Ok(encoder)
+    }
+
+    /// Creates a brotli encoder using the specified settings.
+    ///
+    /// # Errors
+    ///
+    /// If any of the preconditions of the parameters are violated, an error is returned.
+    #[doc(alias = "BrotliEncoderSetParameter")]
+    pub fn build_in<A>(&self, alloc: A) -> Result<BrotliEncoder, SetParameterError>
+    where
+        A: GlobalAlloc + 'static,
+    {
+        let mut encoder = BrotliEncoder::new_in(alloc);
+
+        self.configure(&mut encoder)?;
+
+        Ok(encoder)
+    }
+
+    fn configure(&self, encoder: &mut BrotliEncoder) -> Result<(), SetParameterError> {
         if let Some(mode) = self.mode {
             let key = BrotliEncoderParameter_BROTLI_PARAM_MODE;
             let value = mode as u32;
@@ -466,7 +527,7 @@ impl BrotliEncoderOptions {
             encoder.set_param(key, value)?;
         }
 
-        Ok(encoder)
+        Ok(())
     }
 }
 
@@ -551,6 +612,22 @@ impl<R: BufRead> CompressorReader<R> {
         }
     }
 
+    /// Creates a new `CompressorReader<R>` with a newly created encoder using allocator `alloc`.
+    ///
+    /// # Panics
+    ///
+    /// Panics if the encoder fails to be allocated or initialized
+    pub fn new_in<A>(inner: R, alloc: A) -> Self
+    where
+        A: GlobalAlloc + 'static,
+    {
+        CompressorReader {
+            inner,
+            encoder: BrotliEncoder::new_in(alloc),
+            op: BrotliOperation::Process,
+        }
+    }
+
     /// Creates a new `CompressorReader<R>` with a specified encoder.
     ///
     /// # Examples
@@ -598,7 +675,7 @@ impl<R: BufRead> CompressorReader<R> {
         } else {
             Err(IntoInnerError::new(
                 self,
-                io::Error::from(io::ErrorKind::UnexpectedEof),
+                io::ErrorKind::UnexpectedEof.into(),
             ))
         }
     }
@@ -685,7 +762,7 @@ impl<W: Write> CompressorWriter<W> {
     /// # Panics
     ///
     /// Panics if the encoder fails to be allocated or initialized
-    pub fn new(inner: W) -> CompressorWriter<W> {
+    pub fn new(inner: W) -> Self {
         CompressorWriter {
             inner,
             encoder: BrotliEncoder::new(),
@@ -693,6 +770,22 @@ impl<W: Write> CompressorWriter<W> {
         }
     }
 
+    /// Creates a new `CompressorWriter<W>` with a newly created encoder using allocator `alloc`.
+    ///
+    /// # Panics
+    ///
+    /// Panics if the encoder fails to be allocated or initialized
+    pub fn new_in<A>(inner: W, alloc: A) -> Self
+    where
+        A: GlobalAlloc + 'static,
+    {
+        CompressorWriter {
+            inner,
+            encoder: BrotliEncoder::new_in(alloc),
+            panicked: false,
+        }
+    }
+
     /// Creates a new `CompressorWriter<W>` with a specified encoder.
     ///
     /// # Examples
