@@ -127,9 +127,9 @@ pub use decode::DecompressorReader;
 pub use decode::DecompressorWriter;
 
 use brotlic_sys::*;
-use std::io::ErrorKind;
-use std::os::raw::c_int;
-use std::{fmt, io};
+use std::os::raw::{c_int, c_void};
+use std::{fmt, io, ptr};
+use std::alloc::{GlobalAlloc, Layout};
 use std::error::Error;
 
 /// Quality level of the brotli compression
@@ -745,7 +745,7 @@ impl Error for CompressError {}
 
 impl From<CompressError> for io::Error {
     fn from(err: CompressError) -> Self {
-        io::Error::new(ErrorKind::Other, err)
+        io::Error::new(io::ErrorKind::Other, err)
     }
 }
 
@@ -763,7 +763,7 @@ impl Error for DecompressError {}
 
 impl From<DecompressError> for io::Error {
     fn from(err: DecompressError) -> Self {
-        io::Error::new(ErrorKind::Other, err)
+        io::Error::new(io::ErrorKind::Other, err)
     }
 }
 
@@ -978,7 +978,7 @@ impl<I> IntoInnerError<I> {
     }
 
     /// Returns the error which caused the `into_inner` call to fail. This is used to obtain
-    /// ownership of the error in contrast to [`error`].
+    /// ownership of the error in contrast to `error`.
     pub fn into_error(self) -> io::Error {
         self.1
     }
@@ -996,10 +996,44 @@ impl<I> From<IntoInnerError<I>> for io::Error {
     }
 }
 
-impl<I: fmt::Debug> Error for IntoInnerError<I> {}
+impl<I: fmt::Debug + Send> Error for IntoInnerError<I> {}
 
 impl<I> fmt::Display for IntoInnerError<I> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         self.error().fmt(f)
     }
 }
+
+const MIN_ALIGN: usize = 16;
+
+extern "C" fn malloc(opaque: *mut c_void, size: usize) -> *mut c_void {
+    let global_alloc = opaque as *const Box<dyn GlobalAlloc>;
+
+    if size > usize::MAX - 2 * MIN_ALIGN + 1 {
+        return ptr::null_mut();
+    }
+
+    unsafe {
+        let layout = Layout::from_size_align_unchecked(size + MIN_ALIGN, MIN_ALIGN);
+        let alloc = (*global_alloc).alloc(layout);
+        (alloc as *mut usize).write(size);
+
+        (alloc.add(MIN_ALIGN)) as _
+    }
+}
+
+extern "C" fn free(opaque: *mut c_void, address: *mut c_void) {
+    if address.is_null() {
+        return;
+    }
+
+    let global_alloc = opaque as *const Box<dyn GlobalAlloc>;
+
+    unsafe {
+        let alloc = (address as *mut u8).sub(MIN_ALIGN);
+        let size = (alloc as *const usize).read();
+        let layout = Layout::from_size_align_unchecked(size + MIN_ALIGN, MIN_ALIGN);
+
+        (*global_alloc).dealloc(alloc, layout);
+    }
+}
\ No newline at end of file
